searchNodes=[{"ref":"Knigge.html","title":"Knigge","type":"module","doc":"An opinionated way of dealing with behaviours. Opinionated means that it offers an easy way of defining a &quot;facade&quot; for a behaviour. This facade then delegates calls to the real implementation, which is either given directly to Knigge or fetched from the configuration. Knigge can be used directly in a behaviour, or in a separate module by passing the behaviour which should be &quot;facaded&quot; as an option. Motivation Knigge was born out of a desire to standardize dealing with behaviours and their implementations. As great fans of mox we longed for an easy way to swap out implementations from the configuration which lead us to introduce a facade pattern, where a module&#39;s sole responsibility was loading the correct implementation and delegating calls. This pattern turned out to be very flexible and useful but required a fair bit of boilerplate code. Knigge was born out of an attempt to reduce this boilerplate to the absolute minimum. You can read about our motivation in depth in our devblog. Examples Imagine a behaviour looking like this: defmodule MyGreatBehaviour do @callback my_great_callback(my_argument :: any()) :: any() end Now imagine you want to delegate calls to this behaviour like this: defmodule MyGreatBehaviourFacade do @behaviour MyGreatBehaviour @implementation Application.fetch_env!(:my_application, __MODULE__) defdelegate my_great_callback, to: @implementation end With this in place you can simply reference the &quot;real implementation&quot; by calling functions on your facade: MyGreatBehaviourFacade.my_great_callback(:with_some_argument) Knigge allows you to reduce this boilerplate to the absolute minimum: defmodule MyGreatBehaviourFacade do use Knigge, behaviour: MyGreatBehaviour, otp_app: :my_application end Under the hood this compiles down to the explicit delegation visible on the top. In case you don&#39;t want to fetch your implementation from the configuration, Knigge also allows you to explicitely pass the implementation of the behaviour with the aptly named key implementation: defmodule MyGreatBehaviourFacade do use Knigge, behaviour: MyGreatBehaviour, implementation: MyGreatImplementation end defdefault - Fallback implementations for optional callbacks Now imagine you have a more sophisticated behaviour with some optional callbacks: defmodule MySophisticatedBehaviour do @callback an_optional_callback() :: any() @callback a_required_callback() :: any() @optional_callbacks an_optional_callback: 0 end As you would expect Knigge delegates calls to this callback as usual. But since it&#39;s optional this delegation might fail. A common pattern is to check if the implementation exports the function in question: if function_exported?(MyImplementation, :an_optional_callback, 0) do MyImplementation.an_optional_callback() else :my_fallback_implementation end Knigge offers an easy way to specify these fallback implementations with defdefault: defmodule MySophisticatedFacade do use Knigge, behaviour: MySophisticatedBehaviour, otp_app: :my_application defdefault an_optional_callback do :my_fallback_implementation end end Knigge tries to determine at compile-time if the implementation exports the function in question and only uses the default if this is not the case. As such defdefault incurs no runtime overhead and compiles to a simple def. Of course defdefaults can accept arguments as any usual function: defdefault my_optional_callback_with_arguments(first_argument, another_argument) do case first_argument do # ... end end Options Knigge expects either the otp_app key or the implementation key. If neither is provided an error will be raised at compile time. When using the otp_app configuration you can also pass config_key, which results in a call looking like this: Application.fetch_env!(otp_app, config_key). config_key defaults to __MODULE__. By default Knigge does as much work as possible at compile time. This will be fine most of the time. In case you want to swap out the implementation at runtime - by calling Application.put_env/2 - you can force Knigge to do all delegation at runtime. As you might expect this incurs runtime overhead, since the implementing module will have to be loaded for each call. If you want to do delegation at runtime simply pass delegate_at: :runtime as option. For further information about options check the Knigge.Options module."},{"ref":"Knigge.html#options!/1","title":"Knigge.options!/1","type":"function","doc":"Access the options passed to Knigge for a module"},{"ref":"Knigge.html#t:key/0","title":"Knigge.key/0","type":"type","doc":""},{"ref":"Knigge.Behaviour.html","title":"Knigge.Behaviour","type":"module","doc":"Internal module to work with the supplied behaviour. Can be used to fetch the behaviour from the Knigge.Options-struct and to access the callbacks defined by the behaviour. It works whether the module is already compiled or still open, for which it uses Module.open?/1. If the module is still open it directly accesses the module attribute :callback by calling Module.get_attribute/2 and then using Knigge.AST.function_spec_from_callback/1 to transform the AST into a function spec, such as {:my_function, 2} which is being returned as a Keyword list."},{"ref":"Knigge.Behaviour.html#callbacks/1","title":"Knigge.Behaviour.callbacks/1","type":"function","doc":""},{"ref":"Knigge.Behaviour.html#fetch!/1","title":"Knigge.Behaviour.fetch!/1","type":"function","doc":""},{"ref":"Knigge.Behaviour.html#optional_callbacks/1","title":"Knigge.Behaviour.optional_callbacks/1","type":"function","doc":""},{"ref":"Knigge.Code.html","title":"Knigge.Code","type":"module","doc":"Internal module responsible of injecting the delegations into the calling module. Injects the actual delegations to the implementing module. For this it gets registered as a before_compile-hook from where it fetches all callbacks from the behaviour and generates delegating functions for each callback. It also defines the defdefault macro."},{"ref":"Knigge.Code.html#defdefault/2","title":"Knigge.Code.defdefault/2","type":"macro","doc":""},{"ref":"Knigge.Code.html#generate/2","title":"Knigge.Code.generate/2","type":"function","doc":""},{"ref":"Knigge.Implementation.html","title":"Knigge.Implementation","type":"module","doc":"Internal module to work with the supplied implementation. Fetches the implementing modules based on the Knigge.Options.implementation value. Currently it supports passing the implementation directly or fetching it from the application environment."},{"ref":"Knigge.Implementation.html#fetch!/1","title":"Knigge.Implementation.fetch!/1","type":"function","doc":""},{"ref":"Knigge.Implementation.html#fetch_for!/2","title":"Knigge.Implementation.fetch_for!/2","type":"function","doc":""},{"ref":"Knigge.Options.html","title":"Knigge.Options","type":"module","doc":"Specifies valid Knigge-options and allows to validate and encapsulate the options in a struct. Knigge differentiates between required and optional options: Required Knigge requires a way to determine the implementation to delegate to. As such it requires one of the following options (but not both): implementation directly passes the implementation to delegate to otp_app specifies the application for which the implementation has been configured If both or neither are given an ArgumentError is being raised. Optional These options do not have to be given but control aspects on how Knigge does delegation: behaviour The behaviour for which Knigge should generate delegations. Default: the useing __MODULE__. check_if_exists Controls how Knigge checks if the given modules exist, accepts: a boolean (turn the check fully on or off) one or many environment names (atom or list of atoms) - only checks in the given environments [only: &lt;envs&gt;] - equivalent to the option above [except: &lt;envs&gt;] - only checks if the current environment is not contained in the list Default: [except: :test] config_key The configuration key from which Knigge should fetch the implementation. Is only used when otp_app is passed. Default: the useing __MODULE__ delegate_at An atom defining when delegation should happen, either :compile_time or :runtime. Default: :compile_time do_not_delegate A keyword list defining callbacks for which no delegation should happen. Default: [] warn If set to false this disables all warnings generated by Knigge, use with care. Default: true"},{"ref":"Knigge.Options.html#check_if_exists?/1","title":"Knigge.Options.check_if_exists?/1","type":"function","doc":""},{"ref":"Knigge.Options.html#new/1","title":"Knigge.Options.new/1","type":"function","doc":"Checks the validity of the given opts (validate!/1), applies defaults and puts them into the Knigge.Options-struct."},{"ref":"Knigge.Options.html#validate!/1","title":"Knigge.Options.validate!/1","type":"function","doc":"Validates the options passed to Knigge. It ensures that the required keys are present and that no unknown keys are passed to Knigge which might indicate a spelling error. See the moduledocs for details on required and optional options. Examples iex&gt; Knigge.Options.validate!([1, 2, 3]) ** (ArgumentError) Knigge expects a keyword list as options, instead received: [1, 2, 3] iex&gt; Knigge.Options.validate!([]) ** (ArgumentError) Knigge expects either the :implementation or the :otp_app option but neither was given. iex&gt; Knigge.Options.validate!(implementation: SomeModule) [implementation: SomeModule] iex&gt; Knigge.Options.validate!(otp_app: :knigge) [otp_app: :knigge] iex&gt; Knigge.Options.validate!(otp_app: :knigge, check_if_exists: [:test, :prod]) [otp_app: :knigge, check_if_exists: [:test, :prod]] iex&gt; Knigge.Options.validate!(implementation: SomeModule, otp_app: :knigge) ** (ArgumentError) Knigge expects either the :implementation or the :otp_app option but both were given. iex&gt; Knigge.Options.validate!(otp_app: :knigge, the_answer_to_everything: 42, another_weird_option: 1337) ** (ArgumentError) Knigge received unexpected options: [the_answer_to_everything: 42, another_weird_option: 1337] iex&gt; Knigge.Options.validate!(otp_app: &quot;knigge&quot;) ** (ArgumentError) Knigge received invalid value for `otp_app`. Expected atom but received: &quot;knigge&quot; iex&gt; Knigge.Options.validate!(otp_app: :knigge, delegate_at: :compailtime) ** (ArgumentError) Knigge received invalid value for `delegate_at`. Expected :compile_time or :runtime but received: :compailtime iex&gt; Knigge.Options.validate!(otp_app: :knigge, check_if_exists: &quot;test&quot;) ** (ArgumentError) Knigge received invalid value for `check_if_exists`. Expected boolean or environment (atom or list of atoms) but received: &quot;test&quot;"},{"ref":"Knigge.Options.html#with_defaults/1","title":"Knigge.Options.with_defaults/1","type":"function","doc":"Applies the defaults to the given options: check_if_exists = [except: :test] delegate_at = :compile_time do_not_delegate = [] warn = true"},{"ref":"Knigge.Options.html#t:behaviour/0","title":"Knigge.Options.behaviour/0","type":"type","doc":""},{"ref":"Knigge.Options.html#t:check_if_exists/0","title":"Knigge.Options.check_if_exists/0","type":"type","doc":""},{"ref":"Knigge.Options.html#t:config_key/0","title":"Knigge.Options.config_key/0","type":"type","doc":""},{"ref":"Knigge.Options.html#t:delegate_at/0","title":"Knigge.Options.delegate_at/0","type":"type","doc":""},{"ref":"Knigge.Options.html#t:do_not_delegate/0","title":"Knigge.Options.do_not_delegate/0","type":"type","doc":""},{"ref":"Knigge.Options.html#t:envs/0","title":"Knigge.Options.envs/0","type":"type","doc":""},{"ref":"Knigge.Options.html#t:implementation/0","title":"Knigge.Options.implementation/0","type":"type","doc":""},{"ref":"Knigge.Options.html#t:optional/0","title":"Knigge.Options.optional/0","type":"type","doc":""},{"ref":"Knigge.Options.html#t:otp_app/0","title":"Knigge.Options.otp_app/0","type":"type","doc":""},{"ref":"Knigge.Options.html#t:raw/0","title":"Knigge.Options.raw/0","type":"type","doc":""},{"ref":"Knigge.Options.html#t:required/0","title":"Knigge.Options.required/0","type":"type","doc":""},{"ref":"Knigge.Options.html#t:t/0","title":"Knigge.Options.t/0","type":"type","doc":""},{"ref":"Knigge.Options.html#t:warn/0","title":"Knigge.Options.warn/0","type":"type","doc":""}]